/*
 * Layer Two Tunnelling Protocol Daemon
 * Copyright (C) 1998 Adtran, Inc.
 *
 * Mark Spencer
 *
 * This software is distributed under the terms
 * of the GPL, which you should have received
 * along with this source.
 *
 * Miscellaneous but important functions
 *
 */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#if defined(SOLARIS)
# include <varargs.h>
#endif
#include <netinet/in.h>
#include "l2tp.h"
#include "misc.h"
#include "open_debug.h"
#include "open_path.h"
#include "sysinfo.h"
//@@@ib wnlee 2k2/10/26
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


// added for dialondemand start
#include <sys/socket.h>			// socket()
#include <sys/ioctl.h>			// ioctl()
#include <net/if.h>				// struct ifreq
#ifdef NEW_SOCKET
#include <linux/if_packet.h>	// struct sockaddr_ll
#else
#include <net/if_packet.h> 		// struct sockaddr_pkt
#endif
#include <netinet/in.h>			// htons()

// "struct sockaddr" is defined in linux/socket.h
// "struct sockaddr_in" is defined in netinet/in.h
// "struct in_addr" is defined in netinet/in.h

// select()
#include <sys/time.h>
#include <unistd.h>

// inet_ntoa()
#include <arpa/inet.h>

// "struct ether_header" is defined in net/ethernet.h
#include <net/ethernet.h>

// "struct iphdr" is defined in linux/ip.h
#include <linux/ip.h>
// added for dialondemand end
#include "l2tp.h"

//@@@ie
extern	int 	amit_ui_no;
extern	int		autoreconnect;
//extern int dialondemand(char *interface);
//int		wanl2tp_retry =0;


#if 0
void msg_log (int level, const char *fmt, ...)
{
    char buf[256];
    va_list args;
    va_start (args, fmt);
    vsnprintf (buf, sizeof (buf), fmt, args);
    va_end (args);              //bk - otherwise ppc segfaults
    va_start (args, fmt);       //bk
    vfprintf (stderr, fmt, args);
    fflush (stderr);
    
	//openlog (BINARY, LOG_PID, LOG_DAEMON);
    openlog (BINARY, LOG_ODELAY, LOG_DAEMON);
    syslog (level, "%s", buf);
    va_end (args);
}
#else
void msg_log (int level, const char *fmt, ...)
{
	
}
#endif

void set_error (struct call *c, int error, const char *fmt, ...)
{
    va_list args;
    
	va_start (args, fmt);
    c->error = error;
    c->result = RESULT_ERROR;
    c->needclose = -1;
    vsnprintf (c->errormsg, sizeof (c->errormsg), fmt, args);
    if (c->errormsg[strlen (c->errormsg) - 1] == '\n')
        c->errormsg[strlen (c->errormsg) - 1] = 0;
    va_end (args);
}

struct buffer *new_buf (int size)
{
    struct buffer *b = malloc (sizeof (struct buffer));
    
	if (!b || !size || size < 0)
        return NULL;
    b->rstart = malloc (size);
    
	if (!b->rstart)
    {
        free (b);
        return NULL;
    }
    
	b->start = b->rstart;
    b->rend = (void *) ((unsigned int) b->rstart + size - 1);
    b->len = size;
    b->maxlen = size;
    
	return b;
}

inline void recycle_buf (struct buffer *b)
{
    b->start = b->rstart;
    b->len = b->maxlen;
}

#define bufferDumpWIDTH 16
void bufferDump (char *buf, int buflen)
{
    int i = 0, j = 0;
    /* we need TWO characters to DISPLAY ONE byte */
    unsigned char line[2 * bufferDumpWIDTH + 1], *c;

    for (i = 0; i < buflen / bufferDumpWIDTH; i++)
    {
        c = line;
        for (j = 0; j < bufferDumpWIDTH; j++)
        {
            sprintf (c, "%02x ", (buf[i * bufferDumpWIDTH + j]) & 0xff);
            c++;
            c++;                /* again two characters to display ONE byte */
        }
        *c = '\0';
        msg_log (LOG_WARN, "%s: buflen=%d, buffer[%d]: *%s*\n", __FUNCTION__,
             buflen, i, line);
    }

    c = line;
    for (j = 0; j < buflen % bufferDumpWIDTH; j++)
    {
        sprintf (c, "%02x ",
                 buf[(buflen / bufferDumpWIDTH) * bufferDumpWIDTH +
                     j] & 0xff);
        c++;
        c++;
    }
    if (c != line)
    {
        *c = '\0';
        msg_log (LOG_WARN, "%s:             buffer[%d]: *%s*\n", __FUNCTION__, i, line);
    }
}

void do_packet_dump (struct buffer *buf)
{
    int x;
    unsigned char *c = buf->start;

    printf ("packet dump: \nHEX: { ");
    
	for (x = 0; x < buf->len; x++)
    {
        printf ("%.2X ", *c);
        c++;
    };

    printf ("}\nASCII: { ");
    c = buf->start;
    for (x = 0; x < buf->len; x++)
    {
        if (*c > 31 && *c < 127)
        {
            putchar (*c);
        }
        else
        {
            putchar (' ');
        }
        c++;
    }
    printf ("}\n");
}

inline void swaps (void *buf_v, int len)
{
#ifdef __alpha
    /* Reverse byte order alpha is little endian so lest save a step.
       to make things work out easier */
    int x;
    unsigned char t1;
    unsigned char *tmp = (_u16 *) buf_v;

    for (x = 0; x < len; x += 2)
    {
        t1 = tmp[x];
        tmp[x] = tmp[x + 1];
        tmp[x + 1] = t1;
    }
#else

    /* Reverse byte order (if proper to do so) 
       to make things work out easier */
    int x;
    _u16 *tmp = (_u16 *) buf_v;

    for (x = 0; x < len / 2; x++)
        tmp[x] = ntohs (tmp[x]);
#endif
}

inline void toss (struct buffer *buf)
{
    /*
     * Toss a frame and free up the buffer that contained it
     */

    free (buf->rstart);
    free (buf);
}

inline void safe_copy (char *a, char *b, int size)
{
    /* Copies B into A (assuming A holds MAXSTRLEN bytes)
       safely */
    strncpy (a, b, MIN (size, MAXSTRLEN - 1));
    a[MIN (size, MAXSTRLEN - 1)] = '\000';
}

struct ppp_opts *add_opt (struct ppp_opts *option, char *fmt, ...)
{
    va_list args;
    struct ppp_opts *new, *last;
    
	new = (struct ppp_opts *) malloc (sizeof (struct ppp_opts));
    
	if (!new)
    {
        msg_log (LOG_WARN,
             "%s : Unable to allocate ppp option memory.  Expect a crash\n",
             __FUNCTION__);
        return NULL;
    }
    new->next = NULL;
    va_start (args, fmt);
    vsnprintf (new->option, sizeof (new->option), fmt, args);
    va_end (args);
    if (option)
    {
        last = option;
        while (last->next)
            last = last->next;
        last->next = new;
        return option;
    }
    else
        return new;
}

void opt_destroy (struct ppp_opts *option)
{
    struct ppp_opts *tmp;
    while (option)
    {
        tmp = option->next;
        free (option);
        option = tmp;
    };
}

//@@@ib wnlee 2k2-10-11
int write_csf (int status1, int ui_no, char *remote_ip, int tid, int cid, int type)
{
    /*      Format ==> Status:Remote IP:Tunnel id:Call id
     *
     *       Status
     *         0      failed to connect
     *         1      connected
     *        99      connecting 
     *
     *       Type
     *         0      LNS
     *         1      LAC
     *         2      BOTH (LAC or LNS)
     */
    char 			filename[100];
    int 			size = 0;
    DIR 			*dirh;
    struct dirent 	*dirp;
	FILE 			*fd;

#ifdef _WAN_L2TP
	char			temp[23], *cp, command[132];
#endif

#ifdef FOR_DEBUG
		if( type==LAC )
		{
			printf("l2tpd: write_csf: which:%d\r\n",(ui_no+MAX_NO_ACCOUNT_POOL));
		}
		else
		{
			printf("l2tpd: write_csf: which:%d\r\n",ui_no);
		}
#endif
	
    bzero (filename, 100);
    
        
    if (type == LAC)
    {
        // unlink ui_no-th file
        if ((dirh = opendir (LAC_DIR)) == NULL)
        {
           // Nprintf("opendir [%s] failed\n",LAC_DIR);
            return -1;
        }
        for (dirp = readdir (dirh); dirp != NULL; dirp = readdir (dirh))
        {
            if (!strcmp (dirp->d_name, "."))
                continue;
            if (!strcmp (dirp->d_name, ".."))
                continue;
            size =
                snprintf (filename, 99, "l2tpc%d",
                          (ui_no + MAX_NO_ACCOUNT_POOL));
            *(filename + size) = 0;
            
			Dprintf("l2tpd: write_csf: dirp->d_name [%s] filename [%s]\n\r", dirp->d_name, filename);
            
			if (strcmp (dirp->d_name, filename) == 0)
            {
                size =
                    snprintf (filename, 99, "%s/%s", LAC_DIR, dirp->d_name);
                *(filename + size) = 0;
                unlink (filename);
            }
        }
        closedir(dirh);
		// Prepare new file
        size = snprintf (filename, 99, "%s/l2tpc%d", LAC_DIR, (ui_no + MAX_NO_ACCOUNT_POOL));
        *(filename + size) = 0;
    }
    else if (type == LNS)
    {
        size =
            snprintf (filename, 99, "%s/l2tps%d", LNS_DIR,ui_no);
    	*(filename + size) = 0;
	}

    fd = fopen (filename, "w+");
    if (fd == NULL)
    {
		Dprintf("%s:%s:%d - Can't open file %s\n", __FILE__, __FUNCTION__, __LINE__, filename);
       //Nprintf("can't open %s\n",filename);
        return -1;
    }
    fprintf (fd, "%s\n%d\n%d\n%d\n", remote_ip, tid, cid, status1);
    fclose (fd);
	
#if 0    //move to pppd ( in auth.c)  Mark by Eric -2005/12/12

//#ifdef _WAN_L2TP
    Sprintf(filename,L2TP_WANTYPE_FILE/*"/var/run/l2tpc/l2tp_wantype_info"*/);
	if( (fd=fopen(filename,"r")) != NULL )
	{
		fgets(temp,sizeof(temp),fd);
		fclose(fd);
		Dprintf("l2tpd: write_csf: get WAN type from /var/run/l2tpc/l2tp_wantype_info = %d\n",atoi(temp));
	}
		
	if ((type==LAC) && (atoi(temp)==WT_L2TP ))
    {
    	usleep(300);
		Sprintf(filename,"/var/config/ppp/pppd.%d",MAX_L2TP_CLIENT_NUM-1+MAX_NO_ACCOUNT_POOL);
		fd = fopen (filename, "r");
	    if (fd == NULL)
	    {
			Dprintf("l2tpd: write_csf: can not open file to get ppp string from %s\n",filename);
	    } else {
	    	fgets(temp,sizeof(temp),fd); 
			fgets(temp,sizeof(temp),fd); 
			fclose(fd); 

			ERASE_CARRY(temp);
			Dprintf("l2tpd: write_csf: get ppp string from %s = %s\n",filename, temp);

	    	if (status1 == 0)
	    	{
				//Dprintf("l2tpd: write_csf: Del the DROP PREROUTING rule for PPP WAN --> LAN\r\n");
				//Sprintf(command,"iptables -t nat -D PREROUTING -i %s -o " LANPORTNAME " -j DROP",temp);
				//Sprintf(command,"iptables -t nat -D PREROUTING -i %s -o " BRIDGEPORTNAME " -j DROP",temp);
				//System(command);  //command illegal in PREROUTING NAT table
				//wanl2tp_retry++;
	    		//l2tp_reconnect();
	    	} else if (status1 == 1) {
#if 0
				Dprintf("l2tpd: write_csf: Add the DROP PREROUTING rule for PPP WAN --> LAN\r\n");
				//Sprintf(command,"iptables -t nat -I PREROUTING -i %s -o " LANPORTNAME " -j DROP",temp);
				Sprintf(command,"iptables -t nat -I PREROUTING -i %s -o " BRIDGEPORTNAME " -j DROP",temp);
#else
				Dprintf("l2tpd: write_csf: Run ed 4 background\r\n");
				Sprintf(command,"ed 4&");
				System(command);
#endif
			}
		}
		return 0;
	}
#endif
    
    return 0;
}

int modify_ui_files (int status1,int ui_no,char *remote_ip, int tid, int cid, int action)
{
    /*       Filename format: 
     *       --------------------------
     *         dial in   n-remote_ip:ourtid:ourcid
     *         dial out  remote_ip:ourtid:ourcid
     *
     *       Status of call
     *       ---------------------------
     *         0      failed to connect
     *         1      connected
     *        99      connecting 
     *
     *
     *       Action
     *       ---------------------------------------------------------------
     *         0      find all tid-matching files, write_csf(0) ;we are LNS
     *                i.e. match (remote_ip, tid) from LNS_DIR
     *
     *         1      find all tid-matching files, write_csf(0) ;we are LAC
     *                i.e. match (remote_ip, tid) from LAC_DIR
     *
     *         2      When we received StopCCN, we get the target tunnel tid
     *                but we have to see if this tunnel is dial-in or dial-out.
     *                i.e. match (remote_ip, tid) from LNS_DIR and LAC_DIR
     *
     *         3      case CDN
     *                i.e. match (remote_ip, tid, cid) from both 
     *                     LNS_DIR and LAC_DIR directories
     */
    DIR *dirh;
    struct dirent *dirp;
    char filename[MAXSTRLEN];
    char aim_file[MAXSTRLEN];
    char link_status_file[30];
    int size;
	FILE *fd;
	char	command[100];

#ifdef _NETBIOS_L2TP
	char	remotename[23];
	int		nbisd_fd;
	char	ttt[14];
#endif

#ifdef _WAN_L2TP
	char	temp[23];
	char	subnet[23];
	char	subnetlx[23];
	int		what_WAN_type=0;
	int		connect_type=0;
#endif


	if( ui_no<LNS_UI_NO_BASE )
	{
		Dprintf("l2tpd: modify_ui_files: which(LAC):%d\r\n",(ui_no+MAX_NO_ACCOUNT_POOL));
	}
	else
	{
		Dprintf("l2tpd: modify_ui_files: which(LNS):%d\r\n",ui_no);
	}
	
	bzero (aim_file, MAXSTRLEN);
	bzero (filename, MAXSTRLEN);
	bzero (link_status_file, sizeof(link_status_file) );
	
    
    if (action == 0)
    {
        if ((dirh = opendir(LNS_DIR)) == NULL)
        {
			Dprintf("%s:%s:%d - Can't open DIR %s\r\n", __FILE__, __FUNCTION__, __LINE__, LAC_DIR);
            //Nprintf("opendir [%s] failed\n",LNS_DIR);
            return -1;
        }
        
        size = snprintf(aim_file, MAXSTRLEN - 1, "%s/l2tps%d",LNS_DIR,ui_no);
        *(aim_file+size)=0;
        //size = 0;

        for (dirp = readdir(dirh); dirp != NULL; dirp = readdir(dirh))
        {
            if (!strcmp (dirp->d_name, "."))
                continue;
            if (!strcmp (dirp->d_name, ".."))
                continue;

            size = snprintf (filename, MAXSTRLEN - 1, "%s/%s", LNS_DIR, dirp->d_name);  //getfilename
            *(filename + size) = 0;

			if( strcmp(filename,aim_file )!=0 )
			{
				continue;
			}
            // hit (remote_ip, tid) or (remote_ip,tid,cid)
            unlink (filename);
        }
		closedir(dirh);
    }
    else if (action == 1)
    {
        if ((dirh = opendir (LAC_DIR)) == NULL)
        {
            Dprintf("%s:%s:%d - Can't open DIR %s\r\n", __FILE__, __FUNCTION__, __LINE__, LAC_DIR);
			//Nprintf("%s:opendir [%s] failed\n",LAC_DIR); 
            return -1;
        }
        
        size = snprintf(aim_file, MAXSTRLEN - 1, "%s/l2tpc%d", LAC_DIR, (ui_no+MAX_NO_ACCOUNT_POOL));
        *(aim_file+size)=0;
        //size = 0;
        for (dirp = readdir (dirh); dirp != NULL; dirp = readdir (dirh))
        {
            if (!strcmp (dirp->d_name, "."))
                continue;
            if (!strcmp (dirp->d_name, ".."))
                continue;

            size = snprintf (filename, MAXSTRLEN - 1, "%s/%s", LAC_DIR, dirp->d_name);  //getfilename
            *(filename + size) = 0;

            if( strcmp(filename,aim_file )!=0 )
			{
				continue;
			}

            
            if((fd=fopen(filename, "w")) != NULL)
			{
				fprintf (fd, "%s\n%d\n%d\n%d\n", remote_ip, tid, cid, status1);
				fclose (fd);
			}
            else
			{
				Dprintf("%s:%s:%d - Can't Open file '%s'\r\n", 
					__FILE__, __FUNCTION__, __LINE__, filename);
			}
        }
		closedir(dirh);//tt_lin_error
    }
    else                        //action==2 or action==3
    {
        /*  
         * First, look in dial-in directory, unlink the record if hit.
         */
        
        if ((dirh = opendir (LNS_DIR)) == NULL)
        {
			Dprintf("%s:%s:%d - Can't Open DIR '%s'\r\n", 
				__FILE__, __FUNCTION__, __LINE__, LNS_DIR);
            //Nprintf("opendir [%s] failed\n",LNS_DIR);
            return -1;
        }
        
        size = snprintf(aim_file, MAXSTRLEN - 1,"%s/l2tps%d", LNS_DIR, ui_no);
        *(aim_file+size) = 0;
        //size=0;
        for (dirp = readdir(dirh); dirp != NULL; dirp = readdir(dirh) )
        {
            if (!strcmp (dirp->d_name, "."))
                continue;
            if (!strcmp (dirp->d_name, ".."))
                continue;

            size = snprintf (filename, MAXSTRLEN - 1, "%s/%s", LNS_DIR, dirp->d_name);  //getfilename
            *(filename + size) = 0;

            if( strcmp(filename,aim_file )!=0 )
			{
				continue;
			}

            // hit (remote_ip, tid) or (remote_ip,tid,cid)
           	Dprintf("unlink(%s)\n",filename);
            unlink (filename);
        }
        closedir(dirh);
        
		/*
         * Second, look in dial-out directory, 
         *         modify record content as 0 if hit
         */
        
        if ((dirh = opendir(LAC_DIR)) == NULL)
        {
			Dprintf("%s:%s:%d - Can't Open DIR '%s'\r\n",
					__FILE__, __FUNCTION__, __LINE__, LAC_DIR);

            //Nprintf("opendir [%s] failed\n",LAC_DIR);
            return -1;
        }
        
        size = snprintf(aim_file,MAXSTRLEN - 1,"%s/l2tpc%d",LAC_DIR,(ui_no+MAX_NO_ACCOUNT_POOL));
        *(aim_file+size)=0;
        //size=0;
        for (dirp = readdir(dirh); dirp != NULL; dirp = readdir(dirh))
        {

            if (!strcmp (dirp->d_name, "."))
                continue;
            if (!strcmp (dirp->d_name, ".."))
                continue;

            size = snprintf (filename, MAXSTRLEN - 1, "%s/%s", LAC_DIR, dirp->d_name);
            *(filename + size) = 0;
            
            if( strcmp(filename,aim_file )!=0 )
			{
				continue;
			}

            // hit (remote_ip, tid), modify content
            if( (fd=fopen(filename, "w")) != NULL )
			{
				fprintf (fd, "%s\n%d\n%d\n%d\n", remote_ip, tid, cid, status1);
				fclose (fd);
			}
			else
			{
				Dprintf("%s:%s:%d - Can't Open file '%s'\r\n",
						__FILE__, __FUNCTION__, __LINE__, filename);
			}
        }
		closedir(dirh);
    }
    
    Dprintf("l2tpd: modify_ui_files: status:%d ui_no:%d\r\n",status1,ui_no);

    
#ifdef _NETBIOS_L2TP
    if( (fd=fopen("/var/config/l2tp_netbios","r"))!=NULL  )
    {
    	fclose(fd);
    	if( (status1==0) && (ui_no>=LNS_UI_NO_BASE) )
    	{
	    	Sprintf(filename,"/var/run/l2tps/l2tps%d_remotename",ui_no);
	    	if( (fd=fopen(filename,"r"))!=NULL )
	    	{
	    		fscanf(fd,"%s\n",remotename);
	    		fclose(fd);
	    		Sprintf(filename, "/var/config/%s", remotename );
				
				Dprintf("l2tpd: modify_ui_files: filename:%s\r\n",filename);
	    		
				if( (fd=fopen(filename,"r"))!=NULL )
				{
					fscanf(fd,"%s\n%s\n",subnet,subnetlx);
					fclose(fd);
				}
				else
				{
					Dprintf("%s:%s:%d - Can't open file '%s'\r\n",
						__FILE__, __FUNCTION__, __LINE__, filename);
				}
				Dprintf("l2tpd: modify_ui_files: subnet:%s subnetlx:%s\r\n",subnet,subnetlx);
					
				if( (nbisd_fd=open("/var/run/nbisd.fifo", O_WRONLY/*0x0001*/ | O_NONBLOCK/*0x4000*/)) != NULL)
				{
					Sprintf(ttt,"%s 24 3 ",subnetlx);//tt_lin_check
					write(nbisd_fd, ttt, strlen(ttt));
					close(nbisd_fd);
				}
				else
				{
					Dprintf("%s:%s:%d - Can't Open file '/var/run/nbisd.fifo'\r\n",
						__FILE__, __FUNCTION__, __LINE__ );
				}
	    	}
	    }
    }
#endif
    
    if( (status1==0) && (ui_no>=LNS_UI_NO_BASE) )
    {
    	Sprintf(command,"rm -f /var/run/l2tps/l2tps%d_remotename", ui_no);
    	System(command);
    }
    
#ifdef _WAN_L2TP
    
		Dprintf("l2tpd: modify_ui_files: Now L2TP WAN type compiling switch is enabled, Try to kill pppd\r\n");

		if( (fd=fopen(L2TP_WANTYPE_FILE,"r")) != NULL )
		{
			fgets(temp,sizeof(temp),fd); 
			what_WAN_type = atoi(temp);
			if (what_WAN_type == WT_L2TP)
			{
				fgets(temp,sizeof(temp),fd); 
				fgets(temp,sizeof(temp),fd); 
				fgets(temp,sizeof(temp),fd); 
				fgets(temp,sizeof(temp),fd);
				fgets(temp,sizeof(temp),fd); 
				connect_type = atoi(temp);
			}
			fclose(fd);

			Dprintf("l2tpd: modify_ui_files: Check " L2TP_WANTYPE_FILE ", find WAN type is %d, connect_type=%d\r\n", \
						what_WAN_type,connect_type);
			Dprintf("l2tpd:                  from /var/run/l2tpc/l2tp_wantype_info ..... \r\n");
			System("cat "L2TP_WANTYPE_FILE);
		} else {
			Dprintf("l2tpd: modify_ui_files: Check /var/run/l2tpc/l2tp_wantype_info, but file is not existed\r\n");
			return 0;
		}
		
	    if( status1==0 && ui_no<LNS_UI_NO_BASE )
	    {
	    	if (what_WAN_type == WT_L2TP)
	    	{
	    		Dprintf("l2tpd: modify_ui_files: Now L2TP WAN type, Try to restart l2tpd\r\n");
#if 1
	    			//l2tp_reconnect();
					restart_l2tpd();
		    		return 0;
#else
		    		if ((connect_type == 1) || (connect_type == 2))
		    		{
			    		//wanl2tp_retry++;
						Dprintf("%s,%d connect_type=%d\n",__FILE__,__LINE__,connect_type);
			    		l2tp_reconnect();
			    		return 0;
			    	} else {	//hsujp
						Dprintf("l2tpd: modify_ui_files: Delete the nat PREROUTING dropping ppp WAN to LAN packets\r\n");
						
						Dprintf("l2tpd: modify_ui_files: Add the nat PREROUTING dropping ethernet WAN to LAN packets\r\n");
						
#	if 0	//h command results in the ppp interface to be removed
			    		Sprintf(command,"/var/config/ppp/pppd.%d",(MAX_L2TP_CLIENT_NUM-1+MAX_NO_ACCOUNT_POOL));
						if ( (fd=fopen(command,"r")) != NULL )
						{
							fgets(temp, sizeof(temp), fd);
							fclose(fd);
							
							Sprintf(command,"kill -9 %d",atoi(temp));
							Dprintf("l2tpd: modify_ui_files: Kill pppd: whose pid=%d\r\n",atoi(temp));
							System(command);
						} else {
							Dprintf("l2tpd: modify_ui_files: Kill pppd: Can not find its pid\r\n");
						}
#	endif
						return 0;
			    	}
#endif
	    	}
	    }
#endif
        
    return 0;
}

int write_tid (int tid, int ui_no)
{
    FILE *fd;
    char filename[100];

	Sprintf(filename, "/var/run/l2tpd-%d.tid", (ui_no+MAX_NO_ACCOUNT_POOL));
    fd = fopen (filename, "w+");
    if (fd == NULL)
    {
        //Nprintf("can't open %s\n",filename);
        return -1;
    }
    fprintf (fd, "%d\n", tid);
    fclose (fd);
    return 0;
}

int unlink_ui_files (void)
{
    /*
     *  unlink all l2tpd UI files
     */
    DIR *dirh;
    struct dirent *dirp;
    char filename[100];
    int size = 0;

    if ((dirh = opendir("/var/run")) == NULL)
    {
        //Nprintf("opendir [%s] failed\n", "/var/run");
        return -1;
    }
    for (dirp = readdir(dirh); dirp != NULL; dirp = readdir(dirh))
    {
        if (strstr (dirp->d_name, "l2tpd-") == NULL)
            continue;
        size = snprintf (filename, 99, "%s/%s", "/var/run", dirp->d_name);
        *(filename + size) = 0;

//#ifdef FOR_DEBUG
//        Nprintf("unlink %s\n", filename);
//#endif
        unlink (filename);
    }
    closedir(dirh);
    
	if ((dirh = opendir(LAC_DIR)) == NULL)
    {
        //Nprintf("opendir [%s] failed\n", LAC_DIR);
        return -1;
    }
    for (dirp = readdir(dirh); dirp != NULL; dirp = readdir(dirh))
    {
        if (!strcmp (dirp->d_name, "."))
        	continue;
        if (!strcmp (dirp->d_name, ".."))
        	continue;
		if (!strcmp (dirp->d_name, "l2tp_wantype_info"))	//eric add 
			continue;
		
        size = snprintf(filename, 99, "%s/%s", LAC_DIR, dirp->d_name);
        *(filename + size) = 0;
        unlink (filename);
    }
    closedir(dirh);

    if ((dirh = opendir(LNS_DIR)) == NULL)
    {
        //Nprintf("opendir [%s] failed\n", LNS_DIR);
        return -1;
    }
    for (dirp = readdir(dirh); dirp != NULL; dirp = readdir(dirh))
    {
        if (!strcmp (dirp->d_name, "."))
        	continue;
        if (!strcmp (dirp->d_name, ".."))
        	continue;
		if (!strcmp (dirp->d_name, "l2tp_wantype_info"))
			continue;
		
        size = snprintf(filename, 99, "%s/%s", LNS_DIR, dirp->d_name);
        *(filename + size) = 0;

#ifdef FOR_DEBUG
       	Nprintf("unlink(%s)\n", filename);
#endif
        unlink (filename);
    }
    closedir(dirh);

    return 0;
}

#ifdef _WAN_L2TP
void l2tp_reconnect( void )
{
	unsigned char	command[100];
	FILE	*fd;
	unsigned char	temp[23];
	unsigned char	wtype[23];
	unsigned char	user[23];
	unsigned char	ip[23];
	unsigned char	maxidle[23];
	unsigned char	iptype[23];
	unsigned char	reconnect[23];
	unsigned char	force[23];
	unsigned char	filename[100];
	
	
	
	if( (fd=fopen(L2TP_WANTYPE_FILE,"r")) != NULL )
	{
		memset(wtype,0x00,23);
		memset(user,0x00,23);
		memset(ip,0x00,23);
		memset(maxidle,0x00,23);
		memset(iptype,0x00,23);
		memset(reconnect,0x00,23);
		memset(force,0x00,23);
		fscanf(fd,"%s\n%s\n%s\n%s\n%s\n%s\n%s\n"
				,wtype,user,ip,maxidle,iptype,reconnect,force);
		fclose(fd);

		Warn( (strlen(wtype)>=sizeof(wtype)),			"String 'wrype' Out of bounder");
		Warn( (strlen(user)>=sizeof(user)),				"String 'user' Out of bounder");
		Warn( (strlen(ip)>=sizeof(ip)),					"String 'ip' Out of bounder");
		Warn( (strlen(maxidle)>=sizeof(maxidle)),		"String 'maxidle' Out of bounder");
		Warn( (strlen(iptype)>=sizeof(iptype)),			"String 'iptype' Out of bounder");
		Warn( (strlen(reconnect)>=sizeof(reconnect)),	"String 'reconnect' Out of bounder");
		Warn( (strlen(force)>=sizeof(force)),			"String 'force' Out of bounder");
	}

	Dprintf("l2tpd: reconnect: wtype:%s\r\n",wtype);
	Dprintf("l2tpd: reconnect: user:%s\r\n",user);
	Dprintf("l2tpd: reconnect: ip:%s\r\n",ip);
	Dprintf("l2tpd: reconnect: maxidle:%s\r\n",maxidle);
	Dprintf("l2tpd: reconnect: iptype:%s\r\n",iptype);
	Dprintf("l2tpd: reconnect: reconnect:%s\r\n",reconnect);
	Dprintf("l2tpd: reconnect: force:%s\r\n",force);

	if( (fd=fopen(L2TP_WANTYPE_FILE,"w")) != NULL )
	{
		fprintf(fd,"%s\n%s\n%s\n%s\n%s\n%s\n0\n" 
				,wtype,user,ip,maxidle,iptype,reconnect);
		fclose(fd);
	}
	
	if( atoi(reconnect)==0 )	//wait-for-traffic
	{
		Sprintf(filename, "/var/run/l2tpc/l2tpc%d", (MAX_L2TP_CLIENT_NUM-1+MAX_NO_ACCOUNT_POOL));
		if( (fd=fopen(filename, "w")) != NULL)
		{
			fprintf(fd, "0\n");
			fprintf(fd, "0\n");
			fprintf(fd, "0\n");
			fprintf(fd, "2\n");
			fclose(fd);
		}
		//if( dialondemand(LANPORTNAME)==1 )
		if( dialondemand(BRIDGEPORTNAME)==1 )
		{
			Sprintf(command,"echo \"t %s %d 0 %s\" > /var/run/l2tp-control"
							,user
							,(MAX_L2TP_CLIENT_NUM-1+MAX_NO_ACCOUNT_POOL)
							,ip);
			System(command);
		}
	}
	else if( atoi(reconnect)==1 )	//auto-reconnect
	{
		Sprintf(command,"echo \"t %s %d 1 %s\" > /var/run/l2tp-control"
							,user
							,(MAX_L2TP_CLIENT_NUM-1+MAX_NO_ACCOUNT_POOL)
							,ip);
		System(command);
	}
	if( atoi(reconnect)==2 )	//manuall
	{
#if 0
		Sprintf(filename, "/var/run/l2tpc/l2tpc%d", (MAX_L2TP_CLIENT_NUM-1+MAX_NO_ACCOUNT_POOL));
		if( (fd=fopen(filename, "w")) != NULL)
		{
			fprintf(fd, "0\n");
			fprintf(fd, "0\n");
			fprintf(fd, "0\n");
			fprintf(fd, "2\n");
			fclose(fd);
		}
#endif	
		while(1)
		{
			Dprintf("Wait for client press button!\n");
			if(force==1)
			{
				Sprintf(command,"echo \"t %s %d 0 %s\" > /var/run/l2tp-control"
								,user
								,(MAX_L2TP_CLIENT_NUM-1+MAX_NO_ACCOUNT_POOL)
								,ip);
				System(command);
				break;
			}
			sleep(1);
		}
		
	}
}

int dialondemand(char *interface)
{
	int 					o = 1;
	int 					ethfd;
	struct ifreq 			ifr;
	#ifdef NEW_SOCKET
	  struct sockaddr_ll 	sap;
	#else
	  struct sockaddr_pkt 	sap;
	#endif
	//struct timeval 		tv;
	struct in_addr 			ip, ip2, netmask, netid;
	struct in_addr 			temp_ip;
	unsigned char 			*uptr1, *uptr2;	
	fd_set 					fdread;
	struct 					ether_header *eth_hdr;
	struct iphdr 			*ip_hdr;
	char 					str[100];

	char 					des_ip_str[16], src_ip_str[16];
							//EX: "255:255:255:255\0"
	char					filename[100];
	char					c_temp[23];
	FILE					*fd;
	
	FUNC_LOG

	// 1. --- Check parameters
#if 0		//programmer must be careful of this
	if(strlen(interface) == 0)
	{
		printf("dialondemand: please fill in LAN interface\r\n");
		return -1;
	}
#endif

	// 2. --- Function entrance
	Dprintf("l2tpd: dialondemand: Starting to monitor LAN interface %s.\r\n", interface);

	
	// 3. --- Function initialization
	memset(&ifr, 0, sizeof(struct ifreq));
	memcpy(ifr.ifr_name, interface, strlen(interface));

	// 4. --- Create socket
	Dprintf("l2tpd: dialondemand: Create socket: ");

#ifdef NEW_SOCKET
		ethfd = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
#else
		ethfd = socket(AF_PACKET, SOCK_PACKET, htons(ETH_P_ALL));
#endif

	if(ethfd == -1)
	{
		Dprintf("failed to create socket.\r\n");
		exit(1);
	} else {
		Dprintf("OK!\r\n");
	}

	// 5. --- get hardware address
	Dprintf("l2tpd: dialondemand: Get hardware address: ");
	if(ioctl(ethfd, SIOCGIFHWADDR, &ifr))
	{
		Dprintf("failed to get hardware address.\r\n");
		exit(1);
	} else  {
		Dprintf("%02X:%02X:%02X:%02X:%02X:%02X\r\n", \
				ifr.ifr_hwaddr.sa_data[0], ifr.ifr_hwaddr.sa_data[1], \
				ifr.ifr_hwaddr.sa_data[2], ifr.ifr_hwaddr.sa_data[3], \
				ifr.ifr_hwaddr.sa_data[4], ifr.ifr_hwaddr.sa_data[5] );
	}


	// 6. --- is this interface a ethernet device or not
#if 0
	Dprintf("dhcpcd: Check interface: ");
	if(ifr.ifr_hwaddr.sa_family != ARPHRD_ETHER)
	{
		Dprintf("%s is not an ethernet device.\r\n", ifr.ifr_name);
		exit(1);
	} else {
		Dprintf("%s is an ethernet device.\r\n", ifr.ifr_name);
	}
#endif

	// 7. --- get interface ip
	Dprintf("l2tpd: dialondemand: Get IP address: ");
	if(ioctl(ethfd, SIOCGIFADDR, &ifr) < 0)
	{
		Dprintf("failed to get IP address.\r\n");
		exit(1);
	}
	ip = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr;
	Dprintf("%s\r\n", inet_ntoa(ip));

	// 8. --- get interface netmask
	Dprintf("l2tpd: dialondemand: Get Netmask: ");
	if(ioctl(ethfd, SIOCGIFNETMASK, &ifr) < 0)
	{
		Dprintf("failed to get netmask.\r\n");
		exit(1);
	}
	netmask = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr;
	Dprintf("%s\r\n", inet_ntoa(netmask));

	// 9. --- get network id
	Dprintf("l2tpd: dialondemand: Get netword id: ");
	uptr1 = (unsigned char *)&ip;
	uptr2 = (unsigned char *)&netmask;
	uptr1[0] &= uptr2[0]; 
	uptr1[1] &= uptr2[1];	
	uptr1[2] &= uptr2[2]; 
	uptr1[3] &= uptr2[3];
	memcpy((void *)&netid, uptr1, sizeof(ip));
	Dprintf("%s\r\n", inet_ntoa(netid));

	// 10. --- get or send broadcast packet
	//if(setsockopt(ethfd, SOL_SOCKET, SO_BROADCAST, &o, sizeof(o)) == -1)
	//	exit(1);

	// set flags
	// IFF_UP: interface is up
	// IFF_BROADCAST: broadcast address valid
	// IFF_NOTRAILERS: avoid use of trailers
	// IFF_RUNNING: resource allocated
	//ifr.ifr_flags = IFF_UP | IFF_BROADCAST | IFF_NOTRAILERS | IFF_RUNNING;
	//if(ioctl(ethfd, SIOCSIFFLAGS, &ifr))
	//	exit(1);

	// 11. --- binding
	Dprintf("l2tpd: dialondemand: Bind packet socket to %s: ", ifr.ifr_name);

#ifdef NEW_SOCKET
	memset(&sap, 0, sizeof(sap));
	sap.sll_family = AF_PACKET;
	sap.sll_protocol = htons(ETH_P_ALL);
	//sap.sll_ifindex = 0;	//atoi(interface + 3);
	if (ioctl(ethfd, SIOCGIFINDEX, &ifr) < 0) {
		Dprintf("l2tpd: dialondemand: ioctl(SIOCFIGINDEX): Could not get interface index");
		exit(1);
   	}
    sap.sll_ifindex = ifr.ifr_ifindex;
	if(bind(ethfd, (struct sockaddr *)&sap, sizeof(sap)) == -1)
#else
	memset(&sap, 0, sizeof(sap));
	sap.spkt_family = AF_PACKET;
	memcpy(sap.spkt_device, interface, strlen(interface));
	if(bind(ethfd, (void *)&sap, sizeof(struct sockaddr)) == -1)
#endif
	{
		Dprintf("failed to bind.\r\n");
		exit(1);
	} 
	else 
	{
#ifdef NEW_SOCKET
		Dprintf("OK! and ifindex=%d\r\n",sap.sll_ifindex);
#else
		Dprintf("OK!\r\n");
#endif
	}

	//12. --- Wait for traffic
	Dprintf("l2tpd: dialondemand: Start waiting input packet: \r\n");
		
	//tv.tv_sec = 5; tv.tv_usec = 0;
	for(;;)
	{
		FD_ZERO(&fdread);
		FD_SET(ethfd,&fdread);
		select(ethfd + 1, &fdread, NULL, NULL, NULL);
		
		if(FD_ISSET(ethfd, &fdread))
		{
			o = read(ethfd, (void *)str, sizeof(str)-1);
			eth_hdr = (struct ether_header *)str;
		//	Dprintf("l2tpd: dialondemand: received a packet, type = %04X\r\n", ntohs(eth_hdr->ether_type));
			if(ntohs(eth_hdr->ether_type) == ETHERTYPE_IP)
			{
		//		Dprintf("                     its ether type = ETHERTYPE_IP\r\n");
		//		usleep(300);
				ip_hdr = (struct iphdr *)((char *)eth_hdr + sizeof(struct ether_header));
				ip.s_addr = ip_hdr->daddr; 
				ip2.s_addr = ip_hdr->saddr;
				temp_ip = ip; 
				Strcpy(des_ip_str,inet_ntoa(temp_ip));
				temp_ip = ip2; 
				Strcpy(src_ip_str,inet_ntoa(temp_ip));
		//		Dprintf("                     source ip=%s, destination ip=%s\r\n",src_ip_str,des_ip_str);
				uptr1 = (unsigned char *)&ip; 
				uptr2 = (unsigned char *)&netmask;
				uptr1[0] &= uptr2[0]; 
				uptr1[1] &= uptr2[1]; 
				uptr1[2] &= uptr2[2]; 
				uptr1[3] &= uptr2[3];
				if ((ip.s_addr != netid.s_addr) 					&&
					(ip.s_addr != inet_addr("224.0.0.0"))		 	&&
					(ip_hdr->daddr != inet_addr("255.255.255.255")) &&
					(ip_hdr->daddr != inet_addr("239.255.255.250"))	   )
				{
					Trace("received a packet from LAN %s (%s),and send to WAN (%s)\r\n" \
							, ifr.ifr_name, src_ip_str, des_ip_str);
					//unsigned char protocol_str[5];
					//if (ip_hdr->protocol == 6) {
					//	Strcpy(protocol_str,"TCP");
					//}
					close(ethfd);
					return 1;
				}
			}
		}
	}	// end for loop
	close(ethfd);
	return 0;
}
#endif
