/* pptp.c ... client shell to launch call managers, data handlers, and
 *            the pppd from the command line.
 *            C. Scott Ananian <cananian@alumni.princeton.edu>
 *
 * $Id: pptp.c,v 1.8 2001/03/05 23:54:52 davidm Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#include <sys/wait.h>
#ifdef EMBED
#include <syslog.h>
#endif
#include "pptp_callmgr.h"
#include "pptp_msg.h"
#include "pptp_gre.h"
#include "version.h"
#include "inststr.h"
#include "pty.h"
#include "open_debug.h"
#include "sysinfo.h"

//wait for traffic prototype
extern int dialondemand(char *interface);


#ifdef SUCC
  #define PPTPDEBUG
  /* 20030521 marked by succ for pptp client reliability problem */
  //#define HAVE_FORK
#endif

#ifdef EMBED
#define	pptp_error(a...) syslog(LOG_INFO, ##a)
#else
#define	pptp_error(a...) { fprintf(stderr, ##a); fprintf(stderr, "\n"); }
#endif

#ifdef FOR_DEBUG
	#ifdef PPTPDEBUG
		#define pptp_debug pptp_error
	#else
		#define pptp_debug(a...)
	#endif
#else
	#define pptp_debug(a...)
#endif

/*
 * because we can be run from init,  it seems unwise to just
 * exit quickly for errors that probably won't have resolved
 * themselves immediately,  wait a bit on exit just to slow
 * it all down a little
 */
#define RESPAWN_DELAY	2

#ifndef PPPD_BINARY
  #define PPPD_BINARY "/var/pppd"
#endif

#define PPTP_WANTYPE_FILE	"/var/run/pptpc/pptp_wantype_info"


void get_ip_address(char *name,struct in_addr* retval);
int open_callmgr(struct in_addr inetaddr, int argc,char **argv,char **envp);
void launch_callmgr(struct in_addr inetaddr, int argc,char **argv,char **envp);
void get_call_id(int sock, pid_t gre, pid_t pppd, 
		 u_int16_t *call_id, u_int16_t *peer_call_id);
void launch_pppd(char *ttydev, int argc, char **argv);
//void reset_pppserver_gateway(void);
//extern unsigned long getWANroute(char *p_wanname);
extern int reset_pppserver_gateway(char *wantype_file);


void usage(char *progname) 
{
	fprintf(stderr,
		"%s\n"
		"Usage:\n"
		" %s hostname[,hostname[,...]] [pppd options]\n", version, progname);
	exit(1);
}

static int signaled = 0;

void do_nothing(int sig) { 
    /* do nothing signal handler. Better than SIG_IGN. */
    signaled = 1;
}

void do_alarm(int sig) {
    /* the SIGALRM handler */
}

sigjmp_buf env;
void sighandler(int sig) {
  siglongjmp(env, 1);
}

#ifdef SUCC
static int ui_num;
//static char peerip[20];
int callmgr_sock_copy;
int pty_fd_copy;

/* signal handler for SIGCHLD (occur when pppd(child of pptp) close */
void child_handler (int mysignal)
{  
    pid_t pid;
    int mystatus; 
    FILE *fd;
    char ppp_command[100];
    char pppno_file[30];
    char pppno[23];
    FILE *ppp_fd;
    int index1;
    char command[50];
    char cmpid[10];
    FILE *fp;
    char filename[200], ctemp[20], buf[100];
	unsigned long procip;
	unsigned char *Des_IP;

    pid = waitpid (-1, &mystatus, WNOHANG);

    /* ******************************************* */
    /* succ : connectection failed or disconnected */
    /*        change status to 0                   */
    /* ******************************************* */
    Sprintf(command, "/var/run/pptpc/pptpc%d", ui_num);
    fd = fopen(command, "w");
    fprintf(fd,"%d",0);
    fclose(fd);
    
#if 0 //aren: this code move to pppd term(); Because this function is started when pptp get SIGCHLD of pppd. 
	  //This method is not safe. There are probabilities of making this code not entered. 06/08/25
    /* ******************************************* */
    /*  joe : connectection failed or disconnected */
    /*        clean iptables !                     */
    /* ******************************************* */
    /* added by joe */
	bzero(pppno_file, sizeof(pppno_file));
	Sprintf(pppno_file, "/var/config/ppp/pppd.%d",ui_num);
	if( (ppp_fd=fopen(pppno_file,"r")) != NULL )
	{
		//bzero(pppno,23);
		fgets(pppno, sizeof(pppno), ppp_fd);
		//bzero(pppno,23);
		fgets(pppno, sizeof(pppno), ppp_fd);
		fclose(ppp_fd);
		ERASE_CARRY(pppno);		
		bzero(ppp_command,100);
		Sprintf(ppp_command, "iptables -t nat -D POSTROUTING -o %s -j MASQUERADE", pppno);
		pptp_debug("pptpd: modify_ui_no: pppno: command:%s\n",command);
		System(ppp_command);
    }
#endif
    /* ************************************************* */
    /* remove pptp_pppd(UI).pid (content is pid of pppd) */
    /* ps. create before launch_pppd                     */ 
    /* ************************************************* */
    Sprintf(command, "/var/run/pptpc/pptp_pppd%d.pid", ui_num);
    remove(command);

    /* ***************** */
    /* kill pptp_callmgr */
    /* ***************** */
    Sprintf(command, "/var/run/pptpc/cm%d", ui_num);
    fd = fopen(command, "r");
    fgets(cmpid, 10, fd);
	fclose(fd);
	remove(command);
    Sprintf(command, "kill %d",atoi(cmpid));
    System(command);

    /* ***************************** */
    /* remove pptp_callmgr link file */
    /* ***************************** */
    //Sprintf(command,"/var/run/pptpc/%s",peerip);
    //remove(command);

    /* ************************************************** */
    /* on close, kill all (everyone in our process group) */
    /* ************************************************** */
    if (pty_fd_copy != -1)
    {
		close(pty_fd_copy);
    }
	if (callmgr_sock_copy != -1)
    {
		close(callmgr_sock_copy);
	}

    pptp_debug("PPTP client disconnect ok!\n");

	if( ui_num==( (MAX_PPTP_CLIENT_NUM-1)+MAX_NO_ACCOUNT_POOL + MAX_L2TP_CLIENT_NUM + MAX_L2TP_SERVER_NUM) )  
	{
		Sprintf(command,"echo 30 > /var/run/monitor.fifo");
		System(command);
	}
	
    exit(0);
}

/* signal handler for SIGUSR2 (occur when pppd ipcp up) */
void tunnel_success(int mysignal)
{
	FILE *fd;
	char command[30];

    /* **************************** */
    /* succ : connectection success */
    /*        change status to 1    */
    /* **************************** */   
    Sprintf(command,"/var/run/pptpc/pptpc%d",ui_num);

    fd = fopen(command,"w");
    fprintf(fd,"%d",1);
    fclose(fd);
}
#endif

int main(int argc, char **argv, char **envp) 
{
	struct in_addr inetaddr;
	int callmgr_sock = -1;
	char ptydev[PTYMAX], ttydev[TTYMAX];
	int pty_fd = -1;
	char tmp_str[10];
	
#ifdef _WAN_PPTP
	unsigned char	filename[100];
	unsigned char	wtype[23];
	unsigned char	user[23];
	unsigned char	ip_type[23];
	unsigned char	max_idle[23];
	unsigned char	connect_type[23];	//auto_reconnect change to connect_type by Aren. 060525
	unsigned char	force_connect[23];
	unsigned char	dhcp_state[23];
	unsigned char	ip_str[23];
#endif

	pid_t parent_pid=0, child_pid=0;
	u_int16_t call_id, peer_call_id;
	  
#ifdef SUCC
	int i, cid, pid2;
	FILE * fd;
	char command[50];
	char *new_argv[argc+2];
	int pid=0;

    if(!strcmp(argv[argc-1],"-D"))
		goto daemon;

#ifndef HAVE_FORK 
	switch(cid = vfork())
#else
	switch(cid = fork())
#endif
	{
	case -1: /* error */
#if 0
		signal(SIGUSR1, SIG_DFL);
		pptp_debug("fork failed %s", strerror(errno));
		sleep(RESPAWN_DELAY);
		goto shutdown;   
		break;
#endif
		pptp_debug("fork failed %s", strerror(errno));
		exit(-1);

	case 0: /* child */
		for(i=0; i<argc; i++)
			new_argv[i] = argv[i];
		new_argv[argc] = "-D";
		new_argv[argc+1] = NULL;
		execv(new_argv[0],new_argv);
		break;
    
	default: /* parent */
		exit(0);
	}
#endif

daemon:

#ifdef SUCC
	(void) signal (SIGCHLD, child_handler); 
	(void) signal (SIGUSR2, tunnel_success); 

	/* ************************** */
	/* succ : conectting          */
	/*        change status to 99 */
	/* ************************** */
	ui_num = atoi(argv[argc-2]);
	Sprintf(command, "/var/run/pptpc/pptpc%d", ui_num);
	fd = fopen(command,"w");
	fprintf(fd,"%d",99);
	fclose(fd);

#endif

	//Aren:pptp|ͨSW@/bin/pptp屼N檺pA
	//]屼ۤvAgJpid file 060530
	Sprintf(command, "/var/run/pptpc/pptp_pppd%d.pid",ui_num);
	Dprintf("%s,%d pptp in client mode pid file=%s\n",__FUNCTION__,__LINE__,command);
    if( (fd=fopen(command,"r")) != NULL )
	{
		pid = 0;
	    fscanf(fd, "%d", &pid);
		if(pid !=0)
		{
			Dprintf("%s,%d kill previos pptp(%d)\n",__FUNCTION__,__LINE__,pid);
			kill(pid, SIGKILL);
		}
       	fclose(fd);
	}
    // write pid
	if((fd = fopen(command,"w")) != NULL) 
	{
	       fprintf(fd,"%d",getpid());
        fclose(fd);
	}
    else 
	{
		Dprintf("PPTPd write pid NG\n");
		exit(-1);
	}
#ifdef _WAN_PPTP
	if(ui_num == (MAX_PPTP_CLIENT_NUM-1+MAX_L2TP_CLIENT_NUM+MAX_L2TP_CLIENT_NUM+MAX_NO_ACCOUNT_POOL ) )
	{
		Sprintf(filename, "/var/run/pptpc/pptp_wantype_info" );
		fd=fopen(filename,"r");
		Warn(fd==NULL, "Can't open file '%s'", filename);
		if( fd!=NULL )
		{
			bzero(wtype,sizeof(wtype));
			bzero(user,sizeof(user));
			bzero(ip_str,sizeof(ip_str));
			bzero(max_idle,sizeof(max_idle));
			bzero(ip_type,sizeof(ip_type));
			bzero(connect_type,sizeof(connect_type));
			bzero(force_connect,sizeof(force_connect));

			fscanf(fd,"%s\n%s\n%s\n%s\n%s\n%s\n%s\n"
				,wtype,user,ip_str,max_idle,ip_type,connect_type,force_connect);
			fclose(fd);

			Warn(strlen(wtype)>sizeof(wtype),		"Out of Bonder 'wtype'");
			Warn(strlen(user)>sizeof(user),			"Out of Bonder 'user'");
			Warn(strlen(ip_str)>sizeof(ip_str),		"Out of Bonder 'ip_str'");
			Warn(strlen(max_idle)>sizeof(max_idle), "Out of Bonder 'max_idle'");
			Warn(strlen(ip_type)>sizeof(ip_type),	"Out of Bonder 'ip_type'");
			Warn(strlen(connect_type)>sizeof(connect_type), "Out of Bonder 'connect_type'");
			Warn(strlen(force_connect)>sizeof(force_connect), "Out of Bonder 'force_connect'");
		}
		
		//regen pptp_wantype_info for wait_for_traffic mode ,set force_connect byte 1->0
		fd=fopen(filename,"w");
		Warn(fd==NULL,"Can't open file '%s'", filename);
		if(fd!=NULL)
		{
			//we need to add force_connect.!
			fprintf(fd,"%s\n%s\n%s\n%s\n%s\n%s\n%d\n"
				,wtype,user,ip_str,max_idle,ip_type,connect_type,0);
			fclose(fd);
		}
		
		//Check previous pppd is running or not, if running kill pppd first.
#if 0
		Aren: this code never enter ...so /bin/pptp has not been killed here.
		Sprintf(filename,"/var/config/ppp/pppd.%d",ui_num);
		if((fd=fopen(filename,"r")) != NULL)
		{
			memset(tmp_str,0,sizeof(tmp_str));	
			fgets(tmp_str,sizeof(tmp_str),fd);
			fclose(fd);
			ERASE_CARRY(tmp_str);
			//Sprintf(command,"kill %s",tmp_str);
			//System(command);
			kill(atoi(tmp_str),SIGTERM);
			unlink(filename);
		}
#endif
		
		Dprintf("%s,%d connect_type=%s,%d\n",__FILE__,__LINE__,connect_type,atoi(connect_type));
		if( atoi(wtype) == WT_PPTP )
		{
			reset_pppserver_gateway(PPTP_WANTYPE_FILE);		//set route rule to pptp server

			if( atoi(connect_type)==0 )
			{
				if( atoi(force_connect)==0 )
				{
					//Wait-for-traffic
				    Sprintf(command, "/var/run/pptpc/pptpc%d", 
						(MAX_PPTP_CLIENT_NUM-1+MAX_L2TP_CLIENT_NUM+MAX_L2TP_CLIENT_NUM+MAX_NO_ACCOUNT_POOL));
    				
					fd = fopen(command, "w");
					Warn(fd==NULL, "Can't open file '%s'", command);
    				if(fd!=NULL)
					{
						fprintf(fd,"%d",2);
    					fclose(fd);
					}
					
					//pptp_debug("PPTPC: pptp.c: Listening to eth0\n");
					//dialondemand(WANPORTNAME);
					pptp_debug("%s(%d): Listening on %s\n",__FILE__,__LINE__,BRIDGEPORTNAME);
					dialondemand(BRIDGEPORTNAME);
				}
			}
			else if( atoi(connect_type)==2 )//manuall
			{
				//Aren:Fstatus pageconnect buttonAҥHstatus]0
				Sprintf(command, "/var/run/pptpc/pptpc%d", ui_num);
				fd = fopen(command,"w");
				fprintf(fd,"%d",0);
				fclose(fd);
				for(;;)
				{
					if( atoi(force_connect)==1 )
					{
						break;
					}
					Dprintf("Wait for user press connect button\n");
					sleep(1);
				}
			}
		}
	}
#endif


#ifdef EMBED
	openlog(argv[0],LOG_PID,LOG_USER);
#endif

	if (argc < 2)
		usage(argv[0]);

#ifdef SUCC
	pid2 = setsid();
	chdir("/");
	/*close(0);
	close(1);
	close(2);*/

    /* 20030521 marked by succ, not necessary anymore */
    /* record pptp pid */
    /* 
    Sprintf(command,"/var/run/pptp/pptp%d.pid",ui_num);
    fd = fopen(command,"w");
    fprintf(fd,"%d/n%d",pid2,getpid());
    fclose(fd);
    */
  #endif

	/* ************************************************** */
	/* Step 1: Get IP address for the hostname in argv[1] */
	/* ************************************************** */
	pptp_debug("Step 1: Get IP address for the hostname in argv[1]");
	
	//reset_pppserver_gateway();		//set route rule to pptp server

	get_ip_address(argv[1], &inetaddr);
	
	if(inetaddr.s_addr == 0) 
	{
		pptp_error("get_ip_address failed, exiting");
		sleep(RESPAWN_DELAY);
 
		//eric add change connect status unconnect change status to 0 -20040427 
		Sprintf(command,"/var/run/pptpc/pptpc%d",ui_num);
		fd = fopen(command,"w");
		Warn(fd==NULL, "Can't open file '%s'",command);
		if(fd!=NULL)
		{
			fprintf(fd,"%d",0);
			fclose(fd);			
		}
		sleep(3);
		system("echo 30 > /var/run/monitor.fifo"); 
		exit(-1);
    }

	/* ******************************************* */
	/* Step 2: Open connection to call manager     */
	/*         (Launch call manager if necessary.) */
	/* ******************************************* */
	pptp_debug("Step 2: Open connection to call manager");
	callmgr_sock = open_callmgr(inetaddr, argc,argv,envp);
	if(callmgr_sock < 0)
	{
		pptp_error("Could not open connection to call manager - terminating");
		sleep(RESPAWN_DELAY);
		
		//eric add change connect status unconnect change status to 0 -20040427
		Sprintf(command,"/var/run/pptpc/pptpc%d",ui_num);
		fd = fopen(command,"w");
		
		fprintf(fd,"%d",0);
 		fclose(fd);
  
		exit(1);
  }
  pptp_debug("callmgr opened - fd = %x", callmgr_sock);

#ifdef SUCC
    /* record peerip for remove link file of pptp_callmgr */
    //strcpy(peerip,inet_ntoa(inetaddr));
#endif

	/* ********************************** */
	/* Step 3: Find an open pty/tty pair. */
	/* ********************************** */
	pptp_debug("Step 3: Find an open pty/tty pair.");
	pty_fd = getpseudotty(ttydev, ptydev);
	if (pty_fd < 0) 
	{
		close(callmgr_sock);
		pptp_error("Could not find free pty.");
		sleep(RESPAWN_DELAY);
	
		//eric add change connect status unconnect change status to 0 -20040427
 		Sprintf(command,"/var/run/pptpc/pptpc%d",ui_num);
 		fd = fopen(command,"w");
 		fprintf(fd,"%d",0);
 		fclose(fd);
		exit(1);
	}
	pptp_debug("got a free ttydev");

	/* ********************** */  
	/* Step 4: fork and wait. */
	/* ********************** */
	pptp_debug("Step 4: fork and wait.");
	signal(SIGUSR1, do_nothing); /* don't die */

#ifndef HAVE_FORK
    switch (child_pid = vfork()) {
#else
    switch (child_pid = fork()) {     //##### succ_for_brecis
#endif
	case -1:
		signal(SIGUSR1, SIG_DFL);
		pptp_debug("vfork failed %s", strerror(errno));
		sleep(RESPAWN_DELAY);
		goto shutdown;

	case 0: /* I'm the child! */
		//signal(SIGUSR1, SIG_DFL);

#ifdef SUCC
		/* record pid of pppd for closing connection */
		Sprintf(command,"/var/run/pptpc/pptp_pppd%d.pid",ui_num);
		fd = fopen(command,"w");
		Warn(fd==NULL, "Can't open file '%s'", command);
		if(fd!=NULL)
		{
			fprintf(fd,"%d",getpid());
			fclose(fd);
		}
#endif

		pptp_debug("entered child");
		pptp_debug("callids established..");
		close(callmgr_sock);     

#ifdef SUCC
                                //1 for -D
		launch_pppd(ttydev, argc-2-1, argv+2); /* launch pppd */
#else
		launch_pppd(ttydev, argc-2, argv+2); /* launch pppd */
#endif

		sleep(RESPAWN_DELAY);

		exit(1); /* in case launch_pppd returns */
		break;

	default: /* parent */
    /*
     * There is still a very small race condition here.  If a signal
     * occurs after signaled is checked but before pause is called,
     * things will hang.
     */
#if 0
		if (!signaled) 
		{
			pause(); /* wait for the signal */
		}
		pptp_error("Error %s", strerror(errno));
#endif /*0*/
		break;
	}
 
	get_call_id(callmgr_sock, parent_pid, child_pid, &call_id, &peer_call_id);

	/* ********************************** */
	/* Step 5: Exchange PIDs, get call ID */
	/* ********************************** */
	pptp_debug("Step 5: Exchange PIDs, get call ID");
#if 0
    /* Step 5b: Send signal to wake up pppd task */
    kill(parent_pid, SIGUSR1);
    sleep(2);
#endif /*0*/

	if (sigsetjmp(env, 1)!=0) 
		goto shutdown;
	signal(SIGINT,  sighandler);
	signal(SIGTERM, sighandler);
	signal(SIGKILL, sighandler);

	{
		char buf[128];

		bzero(buf, sizeof(buf) );
		snprintf(buf, sizeof(buf), "pptp: GRE-to-PPP gateway on %s", ptydev);
		inststr(argc,argv,envp, buf);
	}

#ifdef SUCC
	callmgr_sock_copy = callmgr_sock;
	pty_fd_copy = pty_fd;
#endif

	/* ****************************************************************** */
	/* Step 6: Do GRE copy until close. (reliability  problem is here ??) */
	/* ****************************************************************** */ 
	pptp_debug("Step 6: Do GRE copy until close.");
	pptp_gre_copy(peer_call_id, call_id, pty_fd, inetaddr, /*pid2*/getpid());

shutdown:
	/* on close, kill all (everyone in our process group). */
	kill(0, SIGTERM);
	if (pty_fd != -1)
		close(pty_fd);
	if (callmgr_sock != -1)
		close(callmgr_sock);

	pptp_debug("PPTP client disconnect ok!!\n");

	exit(0);
}

/*
 * search through a possible list of ',' seperated ip addresses, try
 * each one,  if it works then use that one
 */
void get_ip_address(char *name,struct in_addr* retval) 
{
	struct sockaddr_in dest;
	int s;
	char *cp, *np;

	retval->s_addr = 0;
	for (cp = name; cp && *cp; cp = np) 
	{

		if ((np = strchr(cp, ',')) != NULL)
    		*np++ = '\0';
		pptp_error("Trying host %s ...", cp);
		if (inet_aton(cp, retval) == 0) 
		{
			struct hostent *host = gethostbyname(cp);
			if (host==NULL) 
			{
				if (h_errno == HOST_NOT_FOUND)
				{
					pptp_error("gethostbyname: HOST NOT FOUND");
				}
				else if (h_errno == NO_ADDRESS)
				{
					pptp_error("gethostbyname: NO IP ADDRESS");
				}
				else
				{
					pptp_error("gethostbyname: name server error");
				}
				continue;
			}
			if (host->h_addrtype != AF_INET) 
			{
				pptp_error("Host has non-internet address");
				continue;
			}
			memcpy(&retval->s_addr, host->h_addr, sizeof(retval->s_addr));
		}

		if (np)
		{
			*(np - 1) = ','; /* put string back how we found it */
		}

		bzero(&dest, sizeof(dest));
		dest.sin_family = AF_INET;
		dest.sin_port   = htons(PPTP_PORT);
		dest.sin_addr   = *retval;
		pptp_debug("socket");
		if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
		{
			pptp_error("Cannot get socket: %s", strerror(errno));
			continue;
		}
		pptp_debug("Connect");
		signal(SIGALRM, do_alarm);
		alarm(15);
		if (connect(s, (struct sockaddr *) &dest, sizeof(dest)) != -1) 
		{
			alarm(0);
			pptp_error("Connect succeeded");
			close(s);
			return;
		}
		alarm(0);
		close(s);
		pptp_error("Connect failed: %s",strerror(errno));
	}
	retval->s_addr = 0;
	return;
}

int open_callmgr(struct in_addr inetaddr, int argc, char **argv, char **envp) 
{
	/* Try to open unix domain socket to call manager. */
	struct sockaddr_un where;
	const int NUM_TRIES = 3;
	int i, fd;

	/* Open socket */
	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) 
	{
		pptp_error("Could not create unix domain socket: %s", strerror(errno));
		return(-1);
	}

	/* Make address */
	where.sun_family = AF_UNIX;
	//#define PPTP_SOCKET_PREFIX "/var/run/pptpc/"
	snprintf(where.sun_path, sizeof(where.sun_path), 
				PPTP_SOCKET_PREFIX "%s", inet_ntoa(inetaddr));

	for (i=0; i<NUM_TRIES; i++) 
	{
		if (connect(fd, (struct sockaddr *) &where, sizeof(where)) < 0) 
		{
			/* couldn't connect.  We'll have to launch this guy. */
			launch_callmgr(inetaddr, argc,argv,envp);
			sleep(3);
		}
		else 
			return fd;
	}
	close(fd);
	pptp_error("Could not launch call manager after %d tries.", i);
	return -1;   /* make gcc happy */
}

void launch_callmgr(struct in_addr inetaddr, int argc,char**argv,char**envp) 
{
	pid_t pid;
	//int status;
	//pptp_callmgr.h:#define PPTP_CALLMGR_BINARY "/bin/pptp_callmgr"
	const char *callmgr = PPTP_CALLMGR_BINARY;

#ifdef SUCC
	char command[30];
	FILE *fd;
#endif

	/* fork and launch call manager process */
#ifndef HAVE_FORK
	switch (pid=vfork()) {
#else
	switch (pid=fork()) {
#endif
	case -1: /* baaad */
		pptp_error("callmgr vfork failed: %s", strerror(errno));
		break;

	case 0: /* child */
	{ 
#if 0
		int callmgr_main(int argc, char**argv, char**envp);
		char *my_argv[2] = { argv[0], inet_ntoa(inetaddr) };
		char buf[128];
		snprintf(buf, sizeof(buf), "pptp: call manager for %s", my_argv[1]);
		inststr(argc,argv,envp,buf);
		exit(callmgr_main(2, my_argv, envp));
#endif

#ifdef SUCC
		/* record pid of pptp_callmgr for kill pptp_callmgr */
		Sprintf(command, "/var/run/pptpc/cm%d", ui_num );
		fd = fopen(command,"w");
		fprintf(fd,"%d",getpid());
		fclose(fd);
#endif
		execlp(callmgr, callmgr, inet_ntoa(inetaddr), NULL);
		pptp_error("execlp() of call manager [%s] failed: %s",  \
					callmgr, strerror(errno));
		exit(1); /* or we trash our parents stack */
    }

	default: /* parent */
#if 0 /* we don't care about status */
		waitpid(pid, &status, 0);
		if (status!=0)
			pptp_error("Call manager exited with error %d", status);
#endif
		break;
	
	}
}

/* XXX need better error checking XXX */
void get_call_id(int sock, pid_t gre, pid_t pppd, 
		 u_int16_t *call_id, u_int16_t *peer_call_id) 
{
	u_int16_t m_call_id, m_peer_call_id;
	/* write pid's to socket */
	/* don't bother with network byte order, because pid's are meaningless
	 * outside the local host.
	 */
  
	write(sock, (char *)&gre, sizeof(gre));
	write(sock, (char *)&pppd, sizeof(pppd));
	pptp_debug("wrote socket information, waiting for read...");
	read(sock,  (char *)&m_call_id, sizeof(m_call_id));
	read(sock,  (char *)&m_peer_call_id, sizeof(m_peer_call_id));
	pptp_debug("Read socket information !!");
	/* XXX FIX ME ... DO ERROR CHECKING & TIME-OUTS XXX */
	*call_id = m_call_id;
	*peer_call_id = m_peer_call_id;
}

void launch_pppd(char *ttydev, int argc, char **argv) 
{
#ifdef _WAN_PPTP
	char *new_argv[argc+16]; /* XXX if not using GCC, hard code a limit here. */
#else
	char *new_argv[argc+10];
#endif
  
	int i;

	//eric add
	FILE *wanfd;	
	char param[6][25];
	char idle_time[25];
	
	wanfd = fopen("/var/run/pptpc/pptp_wantype_info", "r" );
	if( wanfd != NULL )
	{
		for(i=0;i<6;i++)
		{	//i=0->watype ,i=1->user ,i=2->server_ip ,i=3->idle time ,i=4->ip type ,i=5->auto_reconnect
			memset(param[i],0,sizeof(param[i]));
			fgets(param[i], sizeof(param[i]), wanfd);
			ERASE_CARRY(param[i]);
		}
		fclose(wanfd);
	}
	//add end
  
	//pptp.c:#define PPPD_BINARY "pppd"
	new_argv[0] = PPPD_BINARY;
	new_argv[1] = ttydev;
	new_argv[2] = "38400";

	for (i=0; i<argc; i++)
		new_argv[i+3] = argv[i];

	new_argv[i+5] = argv[i];
	new_argv[i+3] = "-S";
	new_argv[i+4] = new_argv[i+2];

#ifdef _WAN_PPTP
	if( atoi(param[0])==WT_PPTP )
	{	 
  		new_argv[i+5]="defaultroute";
		new_argv[i+6]="noauth";  //add by eric  #pptp_client doesn't need peer's authentication. -20040323

		// set PPTP_client idle time add by eric -20040427 
		// 2005/12/01 mofidy if autoreconncet is set don't add idle time parameter.
  		if( (atoi(param[5])==0) && (atoi(param[3])!=0))		
		{
			new_argv[i+7]="idle";
  			new_argv[i+8]=param[3];  
			new_argv[i+9]="usepeerdns";
			new_argv[i+10]="nodeflate";
			new_argv[i+11]="nobsdcomp";
			new_argv[i+12]="lcp-echo-failure";
			new_argv[i+13]="3";
			new_argv[i+14]="lcp-echo-interval";
			new_argv[i+15]="10";
  			new_argv[i+16]=NULL;
  		}
  		else{
			new_argv[i+7]="usepeerdns";
			new_argv[i+8]="nodeflate";
			new_argv[i+9]="nobsdcomp";
			new_argv[i+10]="lcp-echo-failure";
			new_argv[i+11]="3";
			new_argv[i+12]="lcp-echo-interval";
			new_argv[i+13]="10";
			new_argv[i+14]=NULL;
  		}
	}
	else
	{
		new_argv[i+5] = "noauth";      //add by eric  #pptp_client doesn't need peer's authentication -20040323
		new_argv[i+6] = NULL;
	}
#else
	if( atoi(wantype)!=WT_PPTP )
	{	   
  		new_argv[i+5]="noauth";      //add by eric  #pptp_client doesn't need peer's authentication -20040323
  		new_argv[i+6]=NULL;
	}
#endif

#ifdef FOR_DEBUG 
	for(i=0;new_argv[i]!=NULL;i++)
	{
		syslog(LOG_INFO,"new_argv[%d]=%s",i,new_argv[i]);
	}
#endif 
  
	execvp(new_argv[0], new_argv);
}


